/*
 * Copyright 2004-2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.hdiv.util;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;

import org.apache.taglibs.standard.tag.common.core.ImportSupport;
import org.hdiv.dataComposer.IDataComposer;
import org.hdiv.util.HDIVUtil;

/**
 * General purpose utility methods related to processing a servlet request.
 * 
 * @author Gorka Vicente
 * @since HDIV 2.0
 */
public class JSTLRequestUtilsHDIV {

	/**
	 * By default clients can send the key of a flow execution to be resumed
	 * using an argument with this name ("_flowExecutionKey").
	 */
	private static final String FLOW_EXECUTION_KEY_ARGUMENT_NAME = "_flowExecutionKey";

	/**
	 * Adds HDIV state as a parameter if <code>url</code> references our
	 * application.
	 * 
	 * @param request HTTP request
	 * @param url URL
	 * @return <code>url</code> with the HDIV state added as a new parameter
	 */
	public static String addHDIVParameterIfNecessary(HttpServletRequest request, String url) {

		if (!ImportSupport.isAbsoluteUrl(url)) {

			url = JSTLRequestUtilsHDIV.getContextRelativePath(request, url);
			if (url.indexOf(request.getContextPath()) != -1) {

				url = JSTLRequestUtilsHDIV.composeURL(request, url);
			}

		} else { // URL is absolute

			if (url.indexOf(request.getContextPath()) != -1) {

				url = url.substring(url.indexOf(request.getContextPath()));
				url = JSTLRequestUtilsHDIV.composeURL(request, url);
			}
		}
		return url;
	}

	/**
	 * It creates a new state to store all the parameters and values of the
	 * <code>request</code> and it generates a new encoded values for the
	 * <code>request</code> parameters and adds the HDIV parameter.
	 * 
	 * @param request HTTP request
	 * @param finalLocation the location to redirect to
	 * @return URL with encoded parameters and HDIV parameter
	 */
	public static String composeURL(HttpServletRequest request, String finalLocation) {

		String encodedURL = finalLocation;

		IDataComposer dataComposer = (IDataComposer) request.getAttribute("dataComposer");
		dataComposer.beginRequest(HDIVUtil.getActionMappingName(finalLocation));

		int question = finalLocation.indexOf("?");
		if (question > 0) {

			// generate a new encoded values for the url parameters
			encodedURL = JSTLRequestUtilsHDIV.composeAction(finalLocation, question, "UTF-8", dataComposer);
		}

		return JSTLRequestUtilsHDIV.addExtraParameters(request, dataComposer, encodedURL);
	}

	/**
	 * Adds webflow (_flowExecutionKey) and HDIV (_HDIV_STATE_) parameters.
	 * 
	 * Adds the HDIV parameter, depending on the strategy defined by the user,
	 * to validate the request <code>encodedURL</code>.
	 * 
	 * @param request HTTP request
	 * @param dataComposer HDIV's data composer
	 * @param encodedURL URL encoded
	 * @return <code>url</code> with the HDIV state and _flowExecutionKey
	 *         added as a new parameters.
	 * @see org.hdiv.composer.IDataComposer
	 * @since HDIV 2.0.3
	 */
	public static String addExtraParameters(HttpServletRequest request, IDataComposer dataComposer, String encodedURL) {

		encodedURL = JSTLRequestUtilsHDIV.addFlowExecutionParameter(request, dataComposer, encodedURL);

		return JSTLRequestUtilsHDIV.addHDIVState(dataComposer, encodedURL);
	}

	/**
	 * Adds flow execution key (_flowExecutionKey) as <code>url</code>
	 * parameter if the attribute <code>flowExecutionKey</code> exists in the 
	 * <code>request</code>.
	 * 
	 * @param request HTTP request
	 * @param dataComposer HDIV's data composer
	 * @param url url
	 * 
	 * @returns url with _flowExecutionKey parameter
	 * @since HDIV 2.0.3
	 */
	public static String addFlowExecutionParameter(HttpServletRequest request, IDataComposer dataComposer, String url) {

		String flowExecutionKey = (String) request.getAttribute("flowExecutionKey");
		if (flowExecutionKey == null) {
			return url;
		}

		String composedValue = dataComposer.compose(FLOW_EXECUTION_KEY_ARGUMENT_NAME, flowExecutionKey, false);

		// we check if the url contains parameters
		String separator = (url.indexOf("?") > 0) ? "&" : "?";
		String urlWithFlowExecutionKey = url + separator + FLOW_EXECUTION_KEY_ARGUMENT_NAME + "=" + composedValue;

		return urlWithFlowExecutionKey;
	}

	/**
	 * Adds the HDIV parameter, depending on the strategy defined by the user,
	 * to validate the request <code>encodedURL</code>.
	 * 
	 * @param dataComposer HDIV's data composer
	 * @param encodedURL URL encoded
	 * @return <code>url</code> with the HDIV state added as a new parameter
	 * @see org.hdiv.composer.IDataComposer
	 */
	public static String addHDIVState(IDataComposer dataComposer, String encodedURL) {

		String hdivParameter = (String) HDIVUtil.getHttpSession().getAttribute("HDIVParameter");

		String separator = "";
		String requestId = dataComposer.endRequest();

		if ((requestId.length() <= 0) || (encodedURL.startsWith("javascript:"))) {
			return encodedURL;
		}

		// we check if the url contains parameters
		separator = (encodedURL.indexOf("?") > 0) ? "&" : "?";

		hdivParameter = separator + hdivParameter + "=" + requestId;

		return encodedURL + hdivParameter;
	}

	/**
	 * <p>
	 * It generates a new encoded values for the <code>url</code> parameters.
	 * </p>
	 * <p>
	 * The returned values guarantees the confidentiality in the encoded and
	 * memory strategies if confidentiality indicator defined by user is true.
	 * </p>
	 * 
	 * @param url request url
	 * @param questionIndex index of the first question occurrence in
	 *            <code>url</code> string
	 * @param charEncoding character encoding
	 * @return url with encoded values
	 */
	public static String composeAction(String url, int questionIndex, String charEncoding, IDataComposer dataComposer) {

		String value = url;

		value = value.substring(questionIndex + 1);
		value = value.replaceAll("&amp;", "&");

		String token = null;
		String urlAction = HDIVUtil.getActionMappingName(url);

		StringTokenizer st = new StringTokenizer(value, "&");
		while (st.hasMoreTokens()) {

			token = st.nextToken();
			String param = token.substring(0, token.indexOf("="));
			String val = token.substring(token.indexOf("=") + 1);

			String encodedValue = dataComposer.compose(urlAction, param, val, false, true, charEncoding);
			value = value.replaceFirst(HDIVUtil.protectCharacters(token), param + "=" + encodedValue);
		}
		return url.substring(0, questionIndex + 1) + value;
	}

	/**
	 * Checks if <code>path</code> has an action extension or a jsp page
	 * extension.
	 * 
	 * @param path path
	 * @param extensions extensions table
	 * @return True if <code>path</code> is an action or references a jsp
	 *         page.
	 */
	public static boolean hasActionOrServletExtension(String path, Hashtable extensions) {

		if (path.indexOf("?") > 0) {
			path = path.substring(0, path.indexOf("?"));
		}

		if (path.charAt(path.length() - 1) == '/') {
			return true;
		}

		int pound = path.indexOf("#");
		if (pound >= 0) {
			path = path.substring(0, pound);
		}

		// strip a servlet session ID from
		path = HDIVUtil.stripSession(path);

		if (path.endsWith(".jsp")) {
			return true;
		}

		if (extensions != null) {

			for (Enumeration extensionsIds = extensions.elements(); extensionsIds.hasMoreElements();) {

				Pattern extensionPattern = (Pattern) extensionsIds.nextElement();
				Matcher m = extensionPattern.matcher(path);

				if (m.matches()) {
					return true;
				}
			}
		}

		return (!path.startsWith("/")) && (path.indexOf(".") == -1);
	}

	public static String getContextRelativePath(ServletRequest request, String relativePath) {

		String returnValue = null;

		if (relativePath.startsWith("/")) {
			returnValue = relativePath;
		} else if (!(request instanceof HttpServletRequest)) {
			returnValue = relativePath;
		} else { // relative path
			HttpServletRequest hrequest = (HttpServletRequest) request;
			returnValue = hrequest.getContextPath() + "/" + relativePath;
		}

		return removeRelativePaths(returnValue);
	}

	/**
	 * Deletes from <code>url</code> references to relative directories.
	 * 
	 * @param url url
	 * @return Returns the url <code>url</code> without references to relative
	 *         directories.
	 * @since HDIV 2.0.3
	 */
	public static String removeRelativePaths(String url) {

		String urlWithoutRelativePath = url;

		// .. is illegal in an absolute path according to the Servlet Spec and
		// will cause known problems on Orion application servers.
		if (url.indexOf("..") != -1) {
			Stack stack = new Stack();
			StringTokenizer pathParts = new StringTokenizer(urlWithoutRelativePath.replace('\\', '/'), "/");

			while (pathParts.hasMoreTokens()) {
				String part = pathParts.nextToken();

				if (!part.equals(".")) {
					if (part.equals("..")) {
						stack.pop();
					} else {
						stack.push(part);
					}
				}
			}

			StringBuffer flatPathBuffer = new StringBuffer();
			for (int i = 0; i < stack.size(); i++) {
				flatPathBuffer.append("/").append(stack.elementAt(i));
			}

			urlWithoutRelativePath = flatPathBuffer.toString();
		}

		return urlWithoutRelativePath;
	}

}
